import logging
import os
import toml
import inquirer
from crllm.code.loaders import loaders
from crllm.model.provider import providers


def init_project(path: str):
    if os.path.isfile(os.path.join(path, "crllm_config.toml")):
        logging.error("Config file already exists")
        return

    logging.info("Initializing project in %s", path)

    questions = [
        inquirer.List(
            "provider",
            message="Choose provider",
            choices=providers.keys(),
        ),
        inquirer.List("loader", message="Choose loader", choices=loaders.keys()),
    ]

    answers = inquirer.prompt(questions)

    model = providers[answers["provider"]]()

    model_config = model.get_required_config()

    model_questions = [
        inquirer.Text(key, message=f"Enter {key}") for key in model_config
    ]

    model_answers = inquirer.prompt(model_questions)

    config = {
        "crllm": {
            "provider": answers["provider"],
            "loader": answers["loader"],
        },
        "model": model_answers,
    }

    with open(os.path.join(path, "crllm_config.toml"), "w", encoding="utf8") as file:
        toml.dump(config, file)

    logging.info("Finished.")



from abc import ABC, abstractmethod
import logging
from langchain_community.document_loaders.generic import GenericLoader
from langchain_community.document_loaders.parsers import LanguageParser
from langchain_community.vectorstores import FAISS


from crllm.config.config_service import config_service


class Model(ABC):
    def generate(self, prompt_template, prompt_args):
        config = config_service.get_config()
        model_config = config["model"]
        rag_config = config["rag"]

        model = self._get_model(model_config)

        chain = prompt_template | model

        if config["rag"]["enabled"]:
            self.add_rag_context(prompt_args, rag_config)

            logging.debug(
                prompt_template.format(
                    context=prompt_args["context"], code=prompt_args["code"]
                )
            )

        result = chain.invoke(prompt_args)

        logging.debug(result)
        logging.info(result.usage_metadata)

        return result.content

    def add_rag_context(self, prompt_args, rag_config):
        loader = GenericLoader.from_filesystem(
            path=rag_config["src_path"],
            glob=rag_config["src_glob"],
            parser=LanguageParser(),
            show_progress=True,
        )

        docs = loader.load()

        embedding_config = {"model": rag_config["embedding_model"]}

        vectorstore = FAISS.from_documents(
            documents=docs, embedding=self._get_embeddings(embedding_config)
        )

        def format_docs(docs):
            return "\n\n".join(doc.page_content for doc in docs)

        prompt_args["context"] = format_docs(
            vectorstore.similarity_search(prompt_args["code"], 3)
        )

    @abstractmethod
    def _get_model(self, model_config):
        pass

    @abstractmethod
    def _get_embeddings(self, embeddings_config):
        pass

    @staticmethod
    def get_required_config():
        return []



import toml
from crllm.init_project import init_project


def test_init(tmpdir, monkeypatch):
    file_path = tmpdir.mkdir("test")

    def answer():
        yield {"provider": "ollama", "loader": "file"}
        yield {"model": "llama3"}

    gen = answer()
    monkeypatch.setattr("inquirer.prompt", lambda x: next(gen))

    init_project(str(file_path))

    config_file = toml.load(file_path.join("crllm_config.toml"))

    assert config_file["crllm"]["provider"] == "ollama"
    assert config_file["crllm"]["loader"] == "file"
    assert config_file["model"]["model"] == "llama3"



import os
from langchain_ollama import ChatOllama
from langchain_ollama import OllamaEmbeddings
from crllm.model.model import Model


class OllamaModel(Model):
    def _get_model(self, model_config):
        if os.environ.get("OLLAMA_URL"):
            model_config["base_url"] = os.environ.get("OLLAMA_URL")

        return ChatOllama(**model_config)

    def _get_embeddings(self, embeddings_config):
        return OllamaEmbeddings(**embeddings_config)

    @staticmethod
    def get_required_config():
        return ["model"]



from git import Repo
from crllm.config.config_service import config_service


def git_compare(path):
    repo = Repo(path)

    current_branch = repo.active_branch.name
    config = config_service.get_config()
    main_branch = config["crllm"]["git_main_branch"]

    diff = repo.git.diff(f"{main_branch}...{current_branch}", unified=0)

    code = ""

    for line in diff.splitlines():
        code += line + "\n"

    return code



from crllm.model.provider.azure import AzureModel
from crllm.model.provider.dryrun import DryRunModel
from crllm.model.provider.ollama import OllamaModel
from crllm.model.provider.huggingface import HuggingFaceModel
from crllm.model.provider.openai import OpenAIModel


providers = {
    "ollama": OllamaModel,
    "azure": AzureModel,
    "huggingface": HuggingFaceModel,
    "openai": OpenAIModel,
    "dryrun": DryRunModel,
}



import logging
from git import Repo
from crllm.code.loaders.exceptions.no_code_found import NoCodeFoundException
from crllm.config.config_service import config_service


def git(path):
    config = config_service.get_config()
    use_changed_lines = config["crllm"]["git_changed_lines"]

    repo = Repo(path)
    diff = repo.index.diff(None, create_patch=True)

    if len(diff) == 0:
        raise NoCodeFoundException("No changed files found!")

    code = ""

    if not use_changed_lines:
        changed_files = [item.a_path for item in diff if item.a_path is not None]

        logging.info("Found changed files: %s", changed_files)

        for file in changed_files:
            with open(file, encoding="UTF8") as file:
                code += "\nFile name: " + file + "\n"
                code += file.read()
    else:
        for line in diff:
            code += str(line) + "\n"

    return code



from langchain_openai import AzureChatOpenAI, AzureOpenAIEmbeddings
from crllm.model.model import Model


class AzureModel(Model):
    def _get_model(self, model_config):
        return AzureChatOpenAI(**model_config)

    def _get_embeddings(self, embeddings_config):
        return AzureOpenAIEmbeddings(**embeddings_config)

    @staticmethod
    def get_required_config():
        return ["azure_deployment", "api_version"]



import logging
import os
from deepmerge import always_merger
import toml

from crllm.config import CONFIG_DIR


class ConfigService:
    default_config_path = os.path.join(CONFIG_DIR, "config.toml")
    config_path = "./crllm_config.toml"
    config = None

    def set_config_path(self, path):
        self.config_path = path

    def get_config(self):
        if self.config:
            return self.config

        self.config = toml.load(self.default_config_path)

        if not os.path.isfile(self.config_path):
            return self.config

        project_config = toml.load(self.config_path)
        self.override_config(project_config)

        logging.debug(self.config)

        return self.config

    def override_config(self, config: dict):
        if not self.config:
            self.get_config()

        self.config = always_merger.merge(self.config, config)


config_service = ConfigService()
